export {
    Extension,
    gettext as _
} from 'resource:///org/gnome/shell/extensions/extension.js';

export * as AppFavorites from 'resource:///org/gnome/shell/ui/appFavorites.js';
export * as AltTab from 'resource:///org/gnome/shell/ui/altTab.js';
export * as Main from 'resource:///org/gnome/shell/ui/main.js';
export * as OsdWindow from 'resource:///org/gnome/shell/ui/osdWindow.js';
export * as PanelMenu from 'resource:///org/gnome/shell/ui/panelMenu.js';
export * as PopupMenu from 'resource:///org/gnome/shell/ui/popupMenu.js';
export * as SwitcherPopup from 'resource:///org/gnome/shell/ui/switcherPopup.js';
export * as WindowManager from 'resource:///org/gnome/shell/ui/windowManager.js';
